/**************************
  LISSAJOU.C
      This program displays spherical Lissajous figures
      and writes a data file in either Vivid, Persistence of Vision
      (PoV), or Connect the Dots Smoother (CTDS) format.

      Program originally written by Dan Farmer using algorithms from
      Clifford Pickover.  Converted from QuickBasic to C by Aaron C. Caba.
      See Scientific American January 1991 and Omni February 1990 for
      excellent examples by Pickover.

      Compiled with Turbo C++ 1.0 under the LARGE memmory model

History:

 Version 1.51:
    Stuff left to do:
        ESC problems in rotate routine.
        guibolt has never quite worked right.

  04/14/92 ACC fix rotate_fig/disp_spheres/quiksort so they all can rotate
               and display around any axis
               Add axis display on viewing screen

 Version 1.5:

  03/10/92 ACC I decided to change to version number to differentiate
               this from some V. 1.4 betas running around.
               Add the equation display on the main screen

  03/09/92 ACC Finally was able to change to the Large memmory model --
               I found that I was running out of stack space.

 Version 1.4:
  02/11/92 ACC Major rearangement of all code:
               move all field manipulation functions to FIELD.C
               clean up header files so only necessary stuff is in each

   1/07/92 ACC Change all scatered calls to 'exit()' to error_fn.
  12/04/91 ACC Begin converting Dan's original Basic code to C.
               Add Vivid output, CTDS output, change DKB to PoV output.
               Write up some kind of documentation.

 Version 1.3:
  03/08/91 DMF Drop FRACTINT color map option.  Doubt if anyone uses it.
               Make colors all color declarations and use the names in
               the objects, rather than the values.  Easier to edit when
               you have 100s of objects and several colors to change.

  01/16/90 DMF Change most GOSUBS to SUBPROGRAMS and FUNCTIONS
               Specify type of most integers. (Default is single precision)
               Replace Shellsort with Quicksort
               Allow pausing of display, with option to write only the
               spheres displayed.
               Use Fractint color map or specified color(s) for DKB script.

 Revision: DMF Add EXPONENT variable
               Bug Fix: CLOSE #1 when finished with it.
               In WRITE.HEADER, use correct variable for R1 reference.
 *************************/

#include "lissajou.h"

/* ****************************** functions ****************************** */

/* Initalize the 'field' variable with:
                field label
                      length
                      screen (x,y) position
                      type
*/
void initfield(char *data[])
{
    int i,
        rotate_dy=0;

    if(getgraphmode() == VGAHI) rotate_dy=13;


    /* the labels and default data for the data fields */
    for(i=0; i<MAXFIELDS; i++) {
        strcpy(field[i].label,label[i]);
        strcpy(data[i],field_default[i]);
    }
    /* length of each field */
    field[MAIN_RADIUS].len=3;   field[A].len=4;
    field[B].len=4;             field[EXPONENT_X].len=1;
    field[EXPONENT_Y].len=1;    field[EXPONENT_Z].len=1;
    field[SPH_RADIUS].len=2;    field[SPHERES].len=4;
    field[AXIS].len=1;          field[ALGO].len=1;
    field[ROTATE].len=4;        field[FILE_NAME].len=8;
    field[NUMCOLORS].len=2;     field[COLOR_NAME].len=45;
    field[ALGO_EQN].len=1;

    /* set text (x,y) coord's of each field */
    for(i=0; i< PARMFIELDS; i++) {  
        field[i].x=2;
        field[i].y=(i+3);
    }
    field[ROTATE].x=48;     field[ROTATE].y=MESSAGE_Y_POS+rotate_dy;
    field[FILE_NAME].x=37;  field[FILE_NAME].y=2;
    field[NUMCOLORS].x=55;  field[NUMCOLORS].y=2;
    field[COLOR_NAME].x=33; field[COLOR_NAME].y=4;
    field[ALGO_EQN].x=2;    field[ALGO_EQN].y=14;

    /* set what 'type' each field contains */
    for(i=0; i < PARMFIELDS; i++) field[i].type=INTEGER;
    field[A].type=field[B].type=DECIMAL;
    field[AXIS].type=XYZ;
    field[ROTATE].type=SIGNED_INTEGER;
    field[FILE_NAME].type=ALPHA;
    field[NUMCOLORS].type=INTEGER;
    field[COLOR_NAME].type=ALPHA;
} /* end initfield */

extern unsigned _stklen=64000U;     /* This value works, but a smaller 
                                       one might work also -- ACC      */


/* ============================= main() =================================== */
int main(void) {
    int
/*        gdriver=DETECT, gmode, */
/*        gdriver=VGA, gmode=VGAHI, */
        gdriver=EGA, gmode=EGAHI,
        errorcode,
        main_radius,                    /* over-all radius of figure        */
        exponent_x, exponent_y,         /* exponenets for figure generation */
        exponent_z,
        spheres,                        /* number of spheres in figure      */
        algo,                           /* which algorithm to use           */
        t,                              /* loop "time" variable             */
        i,                              /* scratch counter variable         */
        exitflag,                       /* if TRUE, then exit               */
        index[MAXSPHERES],              /* array used to order spheres      */
        done;                           /* flag if function is done         */
    float x,y,z,                        /* coords of current point          */
          a,b;                          /* parameters for figure generation */
    char *data[MAXFIELDS],              /* holds user inputed parameters    */
         ch;                            /* input to choose Output/Rotate/ect. */

    /* queue holds (x,y,z) and scaled radius    */
    struct queue_type queue[MAXSPHERES];

#ifdef DEBUG
    char msg[40];   /* debug var */
    unsigned long debug; /* a debugging variable */

    printf("The stack length is: %u\n",_stklen);
    getch();
#endif

    /* init 'data' strings */
    for(i=0; i<MAXFIELDS; i++) {
        if ((data[i]=malloc((MAX_FIELD_LEN+1)*sizeof(char))) == NULL)
            error_fn(EMEM,"data[] init");
        strcpy(data[i],"");
    }

    /* initalize the graphics system */
    registerbgidriver(EGAVGA_driver);
    initgraph(&gdriver, &gmode, "");    /* initalize EGA graphics           */
    errorcode = graphresult();          /* read result of initgraph call    */
    if (errorcode != grOk)              /* an error occurred                */
        error_fn(EGRAPH,grapherrormsg(errorcode));

    setvisualpage(0);   /* choose graphics page #0 for output */
    setactivepage(0);

    /* initalize the screen coordinates */
    cols=getmaxx(); rows=getmaxy();
    xcenter=cols/2; ycenter=rows/2;
    wintop=16; winbottom=rows-50;
    winleft=xcenter/2+28; winright=cols-10;
    winxcenter=(winright+winleft)/2-winleft;  /* circles drawn relative to (0,0) */
    winycenter=(winbottom+wintop)/2-wintop;   /* of window, not screen */

    /* initalize fields with labels, (x,y) pos, length and type         */
    initfield(data);
    assign_default(data);               /* defaults data for each field */

#ifdef DEBUG
    debug=farcoreleft();
    sprintf(msg,"free far core:  %lu",debug);
    disp_message(msg);
    getch();
    debug=coreleft();
    sprintf(msg,"free core:  %lu",debug);
    disp_message(msg);
    getch();
#endif

    setbkcolor(EGA_LIGHTGRAY);
    disp_bkground();                    /* draw windows, field labels       */
    disp_data(data);                    /* display current data (defaults)  */

    exitflag=edit_parms(data);          /* edit the parameters              */
    while (!exitflag) {
        disp_message("Generating...");
        main_radius=atoi(data[MAIN_RADIUS]);    /* convert data to numbers  */
        a=atof(data[A]);
        b=atof(data[B]);
        exponent_x=atoi(data[EXPONENT_X]);
        exponent_y=atoi(data[EXPONENT_Y]);
        exponent_z=atoi(data[EXPONENT_Z]);
        spheres=atoi(data[SPHERES]);
        algo=atoi(data[ALGO]);

        /* now for the calculations */
        for(t=0; t<spheres; t++) {
            switch(algo) {
                case 1: algo1(t,main_radius,a,b,
                              exponent_x,exponent_y,exponent_z,
                              &x,&y,&z);
                        break;
                case 2: algo2(t,main_radius,a,b,
                              exponent_x,exponent_y,exponent_z,
                              &x,&y,&z);
                        break;
                case 3: algo3(t,main_radius,a,b,
                              exponent_x,exponent_y,exponent_z,
                              &x,&y,&z);
                        break;
                case 4: algo4(t,main_radius,a,b,
                              exponent_x,exponent_y,exponent_z,
                              &x,&y,&z);
                        break;
                case 5: algo5(t,main_radius,a,b,
                              exponent_x,exponent_y,exponent_z,
                              &x,&y,&z);
                        break;
            }
            index[t]=t;
            queue[t].x=x; queue[t].y=y; queue[t].z=z;
        }

        disp_message("Sorting...");
        quicksort(queue, index, 0, spheres-1, data[AXIS]);
        disp_spheres(queue, index, spheres, data);

        /* Allow user to choose what to do next:
                Output a script file
                Rotate the on-screen image
                Change generator variables
                Quit */
        done=FALSE;
        while (!done && !exitflag) {
            disp_message("Finished:  [O]=Output script  [R]=Rotate  [CR]=More  [ESC]=Quit");
            ch=getch();
            switch(ch) {
                case RETURN: done=TRUE; break;
                case 'O':
                case 'o': output(queue,index,spheres,data); break;
                case 'R':
                case 'r':
                    rotate_fig(queue, spheres, data[ROTATE], data[AXIS]);
                    disp_message("Sorting...");
                    quicksort(queue, index, 0, spheres-1, data[AXIS]);
                    disp_spheres(queue, index, spheres, data);
                    break;
                default: if (iscntrl(ch)) {
                             if (ch==ESC) exitflag=verify("Exit now ");
                             while(kbhit()) getch();
                         }
                         break;
            }
        }
        if (!exitflag) exitflag=edit_parms(data); /* edit the parameters    */
    } /* end of while(!exit_flag) */
    closegraph();
    printf("Thanks for using Lissajous Ver. %s\n",VERSION);
    printf("By Aaron Caba and Dan Farmer\n");
    return(0);
} /* end main */


/* Rotate the figure by 'angle_str' degrees */
void rotate_fig(struct queue_type queue[], int num,
                char *angle_str, char *axis)
{
    int i, ret;
    float  a, b,    /* work variables */
          at,bt,
          angle;

    disp_message(field[ROTATE].label);
    parm_field=FALSE;
    ret=fieldinput(ROTATE, angle_str, field[ROTATE].label, 0);
    if (ret==ESC) {
        if (verify(" Quit now ")) {
            clean_up_graph();
            exit(0);
        }
        else return;
    }
    
    disp_message("Rotating...");
    angle=PI/180.0*atof(angle_str);
    if (angle != 0) {
        for (i=0; i<num; i++) {

            switch (axis[0]) {
                case 'X' :
                    a=queue[i].z;  b=queue[i].x;
                    break;
                case 'Y' :
                    a=queue[i].x;  b=queue[i].y;
                    break;
                case 'Z' :
                    a=queue[i].y;  b=queue[i].z;
                    break;
            }

            at=a*cos(angle) - b*sin(angle);
            bt=a*sin(angle) + b*cos(angle);
            a=at;   b=bt;

            switch(axis[0]) {
                case 'X' :
                    queue[i].z=a;  queue[i].x=b;
                    break;
                case 'Y' :
                    queue[i].x=a;  queue[i].y=b;
                    break;
                case 'Z' :
                    queue[i].y=a;  queue[i].z=b;
                    break;
            }
        }
    }
} /* end rotate_fig */



/* Display the sorted spheres in the viewing window, with option
   to pause/abort when a key is pressed */
void disp_spheres(struct queue_type queue[], int index[], 
                  int numspheres, char *data[])
{
    int i, color,
        r, ret=TRUE,
        xscr, yscr;     /* screen coordinates of the sphere */
    char msg[71];

    disp_message("Press any key to pause display...");
    setviewport(winleft+8,wintop+8,winright-8,winbottom-8,CLIP_ON);
    clearviewport();
    i=0;
    while (i<numspheres && ret==TRUE) {
        color=colors[i % maxcolors];

        /* adjust coords for the center of the viewing window */
        if (!strcmp("X",data[AXIS])) {
            xscr=(int)queue[index[i]].y;
            yscr=(int)queue[index[i]].z;
            r=scaled_radius(atoi(data[SPH_RADIUS]),queue[index[i]].x);
        }
        else if (!strcmp("Y",data[AXIS])) {
            xscr=(int)queue[index[i]].z;
            yscr=(int)queue[index[i]].x;
            r=scaled_radius(atoi(data[SPH_RADIUS]),queue[index[i]].y);
        }
        else if (!strcmp("Z",data[AXIS])) {
            xscr=(int)queue[index[i]].x;
            yscr=(int)queue[index[i]].y;
            r=scaled_radius(atoi(data[SPH_RADIUS]),queue[index[i]].z);
        }
        xscr += winxcenter;
        yscr += winycenter;

        /* to reverse the graphics system's upside-down 'up' vector */
        yscr = (winbottom - wintop) - yscr;

        setcolor(color);
        setfillstyle(SOLID_FILL,color);
        fillellipse(xscr,yscr,r,r);
        if (kbhit()) {              /* pause and/or stop displaying spheres */
            while (kbhit()) getch();
            disp_axes(data[AXIS]);
            setviewport(0,0,cols,rows,CLIP_OFF);
            sprintf(msg,"Paused at %i spheres.  Continue display ",i+1);
            ret=verify(msg);
            disp_message("Press any key to pause display...");
            setviewport(winleft+8,wintop+8,winright-8,winbottom-8,CLIP_ON);
        }
        i++;
    }

    /* draw coordinate axes for reference */
    disp_axes(data[AXIS]);
    setviewport(0,0,cols,rows,CLIP_OFF);
} /* end disp_spheres */


/* draws a set of coordinate axes on the display screen */
void disp_axes(char *axis)
{
    int save_color=getcolor();
    char a[2], b[2];

    /* assign and draw the axis labels */
    switch (axis[0]) {
        case 'X':
            strcpy(a,"Z");
            strcpy(b,"Y");
            break;
        case 'Y':
            strcpy(a,"X");
            strcpy(b,"Z");
            break;
        case 'Z':
            strcpy(a,"Y");
            strcpy(b,"X");
            break;
    }
    disp_text(2,0,a,EGA_BLUE);
    disp_text(4,2,b,EGA_BLUE);

    /* draw the axis arrows */
    setcolor(EGA_BLUE);
    line(0,5,5,0);
    line(5,0,10,5);
    line(5,0,5,20);
    line(5,20,25,20);
    line(25,20,20,15);
    line(25,20,20,25);

    /* restore original color */
    setcolor(save_color);
}



/* This isn't as bad as it looks!  It scales the radius of the 
   displayed spheres to simulate perspective.  I put the (float)'s in to
   stop integer math from rounding.
   Returns the scaled radius */
int scaled_radius(int r, int z)
{
    if ((VIEWER-z) <= 0) return(0);
    else return((int)ceil( (float)r * (float)VIEWER / (float)(VIEWER-z) ));
}


/* Make sure all entered data is in the proper format or range, assign
   defaults if data isn't O.K. */
void check_data(char *data[])
{
    int i, tempi;
    double tempf;
    char *endptr;

    for(i=0; i < MAXFIELDS; i++) {
        switch(field[i].type) {
            case DECIMAL:
                tempf=strtod(data[i],&endptr);
                if ((strcmp(endptr,"")) || (tempf == 0.0)) 
                    strcpy(data[i],field_default[i]);
                break;
            case INTEGER:
                switch(i) {
                    case ALGO:
                        tempi=atoi(data[ALGO]);
                        if (tempi > MAXALGO) itoa(MAXALGO,data[ALGO],10);
                        else if (tempi == 0) strcpy(data[i],field_default[i]);
                        break;
                    case SPHERES:
                        tempi=atoi(data[SPHERES]);
                        if (tempi > MAXSPHERES)
                            itoa(MAXSPHERES,data[SPHERES],10);
                        else if (tempi < 1)
                            strcpy(data[i],field_default[i]);
                        break;
                    default:
                        if (atoi(data[i]) == 0) {
                            strcpy(data[i],field_default[i]);
                        }
                        break;
                }
                break;
            case SIGNED_INTEGER:
                if (atoi(data[i])==0)
                    strcpy(data[i],field_default[i]);
                break;
            case XYZ:
                if (strlen(data[i]) == 0) strcpy(data[i],field_default[i]);
                else strcpy(data[i],strupr(data[i]));
                break;
        } /* end switch */
    } /* end for loop */
} /* end check_data */


/* Edit the generation parameters.
   Return [EXIT || NO_EXIT] */
int edit_parms(char *data[])
{
    int fieldnum=0, ret, done, exitflag;

    done=exitflag=FALSE;
    parm_field=TRUE;
    disp_message("Press F10 to restore defaults.");
    while (!exitflag && !done) {
        ret=fieldinput(fieldnum,data[fieldnum],"Press F10 to restore defaults.",0);
        check_data(data);               /* make sure data is valid  */
        disp_data(data);
        switch(ret) {
            case UP: fieldnum--;
                     if (fieldnum < 0) fieldnum=0;
                     break;
            case PGDN: done=TRUE;
                       break;
            case PGUP: fieldnum=0;
                       break;
            case ESC:  if (verify(" Quit now ")) exitflag=TRUE;
                       else disp_message("Press F10 to restore defaults.");
                       break;
            case F10:  fieldnum=0;
					   assign_default(data);
                       disp_data(data);
                       break;
            case DOWN:
            default:   fieldnum++;
                       if (fieldnum < PARMFIELDS) break;
                       else { done=TRUE; break;}
        }
    }
    if (exitflag) return(EXIT);
    return(NO_EXIT);
} /* end edit_parms */


/*  Print current values in all data fields */
void disp_data(char *data[])
{
    #define EQN_COLOR    EGA_BLUE
    int i;
    char x_eqn1[21], x_eqn2[21], y_eqn1[21], y_eqn2[21], z_eqn[21];

    parm_field=TRUE;
    for(i=0; i < PARMFIELDS; i++) disp_field(i,data[i],EGA_DARKGRAY,0);

    /* print out formulas coresponding to algo # */
    switch(atoi(data[ALGO])) {
        case 1:
            strcpy(x_eqn1,"x=R*(sin(A*t)*");
            strcpy(x_eqn2,"  cos(B*t)^X)");
            strcpy(y_eqn1,"y=R*(sin(A*t)*");
            strcpy(y_eqn2,"  sin(B*t)^Y)");
            strcpy(z_eqn,"z=R*(cos(A*t)^Z)");
            break;
        case 2:
            strcpy(x_eqn1,"x=R*(sin(A*t)*");
            strcpy(x_eqn2,"  cos(B*t)^X)");
            strcpy(y_eqn1,"y=R*(cos(A*t)*");
            strcpy(y_eqn2,"  cos(B*t)^Y)");
            strcpy(z_eqn,"z=R*(sin(A*t)^Z)");
            break;
        case 3:
            strcpy(x_eqn1,"x=R*(sin(A*t)*");
            strcpy(x_eqn2,"  sin(B*t)^X)");
            strcpy(y_eqn1,"y=R*(sin(A*t)*");
            strcpy(y_eqn2,"  cos(B*t)^Y)");
            strcpy(z_eqn,"z=R*(sin(A*t)^Z)");
            break;
        case 4:
            strcpy(x_eqn1,"x=R/4*(A*sin(2*");
            strcpy(x_eqn2,"  (t-PI/13))^X)");
            strcpy(y_eqn1,"y=R/4*(-B*");
            strcpy(y_eqn2,"  cos(t)^Y)");
            strcpy(z_eqn,"z=R*(sin(A*t)^Z)");
            break;
        case 5: 
            strcpy(x_eqn1,"x=R*(sin(A*t)*");
            strcpy(x_eqn2,"  (cos(A*t)^X))");
            strcpy(y_eqn1,"y=R*(sin(B*t)*");
            strcpy(y_eqn2,"  (sin(B*t)^Y))");
            strcpy(z_eqn,"z=R*(sin(t)^Z)");
            break;
    }
    erase_text(field[ALGO_EQN].x,field[ALGO_EQN].y+1,20);
    disp_text(field[ALGO_EQN].x, field[ALGO_EQN].y+1, x_eqn1, EQN_COLOR);
    erase_text(field[ALGO_EQN].x,field[ALGO_EQN].y+2,20);
    disp_text(field[ALGO_EQN].x, field[ALGO_EQN].y+2, x_eqn2, EQN_COLOR);

    erase_text(field[ALGO_EQN].x,field[ALGO_EQN].y+3,20);
    disp_text(field[ALGO_EQN].x, field[ALGO_EQN].y+3, y_eqn1, EQN_COLOR);
    erase_text(field[ALGO_EQN].x,field[ALGO_EQN].y+4,20);
    disp_text(field[ALGO_EQN].x, field[ALGO_EQN].y+4, y_eqn2, EQN_COLOR);

    erase_text(field[ALGO_EQN].x,field[ALGO_EQN].y+5,20);
    disp_text(field[ALGO_EQN].x, field[ALGO_EQN].y+5, z_eqn, EQN_COLOR);

    #undef EQN_COLOR
}



/* Draw a gui-like panel from (wleft,wtop)-(wright,wbottom) */
void guipanel(int wleft, int wtop, int wright, int wbottom, int toggle)
{
/* Parameter TOGGLE := -1 FOR INSET, 1 FOR OUTSET                       */
/*                     -2 FOR INSET 2 DEEP, 3 TO OUTSET 3 DEEP, ETC.    */
    int depth,i,inset;

    depth=abs(toggle);
    inset=(toggle <0);

    for(i=0; i<=depth; i++) {
        if (inset) {                        /* INSET PANEL  */
            setcolor(8);
            line(wleft+i,  wtop+i,wleft+i,wbottom-i);       /* left side    */
            line(wleft+i,  wtop+i ,wright-i,wtop+i);        /* top line     */
            setcolor(15);
            line(wleft+i,  wbottom-i,wright-i,wbottom-i);   /* bottom line  */
            line(wright-i, wtop+i,wright-i,wbottom-i);      /* right side   */
        }
        else {                              /* OUTSET PANEL */
            setcolor(15);
            line(wleft+i,  wtop+i,wleft+i,wbottom-i);       /* left side    */
            line(wleft+i,  wtop+i,wright-i,wtop+i);         /* top line     */
            setcolor(8);
            line(wleft+i,  wbottom-i,wright-i,wbottom-i);   /* bottom line  */
            line(wright-i, wtop+i,wright-i,wbottom-i);      /* right side   */
        }
    }
} /* end guipanel */


void guibolt(int x, int y, int r, int toggle)
{
/* Parameter TOGGLE := -1 FOR iNSET, 1 FOR OUTSET                       */
/*                     -2 FOR iNSET 2 DEEP, 3 TO OUTSET 3 DEEP, ETC.    */
    int depth, i, inset;

    depth=abs(toggle);
    inset=(toggle<0);

    for(i=0; i<=depth; i++) {
        if (inset) {
            setcolor(15);
            arc(x, y, 0.5*PI,1.5*PI, r);
            arc(x, y, 0.5*PI,1.5*PI, r);
        }
        else {
            setcolor(8);
            arc(x, y, 0.5*PI,1.5*PI, r);
            arc(x, y, 0.5*PI,1.5*PI, r);
        }
    }
}


/* Draw the gui-pannels, credits, and display the field labels */
void disp_bkground(void)
{
    int color,i,y=22,
        bolt_dy=0, win_dy=0;

    if(getgraphmode() == VGAHI) {
        y += 12;
        bolt_dy=120;
        win_dy=115;
    }

    /* Create a GUI-looking window for the data entry fields. */
    guipanel(8,wintop,winleft-7,winbottom,2);

    /* print the field labels */
    for(i=0; i<PARMFIELDS; i++)
        disp_text(field[i].x, field[i].y, field[i].label, EGA_YELLOW);
    disp_text(field[ALGO_EQN].x,field[ALGO_EQN].y,"Equations:", EGA_MAGENTA);

    /* Panel for credits and logo                           */
    guipanel(16,215+win_dy,winleft-15,winbottom-6,1);
    color=EGA_WHITE;
    /* The raised panel can be printed in from rows 34-40   */
    /* starting at column 3 for a length of 17 Characters.  */
    disp_text(3,y  ,"  3-D Lissajous  ",color);
    disp_text(3,y+1,"    Generator    ",color);
    disp_text(3,y+5,"  By Aaron Caba  ",color);
    disp_text(3,y+6," and Dan Farmer  ",color);

    /* Colored Dots "Logo" */
/*    guibolt( 65,BOLT_HT+bolt_dy,10,1);
    guibolt( 90,BOLT_HT+bolt_dy,10,1);
    guibolt(115,BOLT_HT+bolt_dy,10,1);
*/
    color=EGA_GREEN;
    setcolor(color);    circle( 65,BOLT_HT+bolt_dy,8);
    setfillstyle(SOLID_FILL,color); floodfill(65,BOLT_HT+bolt_dy,color);

    color=EGA_MAGENTA;
    setcolor(color);    circle( 90,BOLT_HT+bolt_dy,8);
    setfillstyle(SOLID_FILL,color); floodfill(90,BOLT_HT+bolt_dy,color);

    color=EGA_BLUE;
    setcolor(color);    circle(115,BOLT_HT+bolt_dy,8);
    setfillstyle(SOLID_FILL,color); floodfill(115,BOLT_HT+bolt_dy,color);

/*    guibolt(10,8,4,1);              /* UPPER LEFT BOLT  */
    guibolt(cols-10,8,4,1);         /* UPPER RIGHT BOLT */
*/
    /* Create panel for messages at screen bottom */
    guipanel(8,winbottom+8,winright,rows-8,2);
/*    guibolt(24,rows-24,4,1);
    guibolt(winright-24,rows-24,4,1);
*/
    /* Create a GUI-looking window for the image display */
	guipanel(winleft,wintop,winright,winbottom,-3);   /* SUNKEN PANEL */
} /* end disp_background */


/* These calculate the (x,y,z) coordinate at time 't' */
void algo1(int t, int r, float a, float b,
           int exponent_x, int exponent_y, int exponent_z,
           float *x, float *y, float *z)
{   
    *x = r*(sin(a*t)* pow(cos(b*t),exponent_x));
    *y = r*(sin(a*t)* pow(sin(b*t),exponent_y));
    *z = r*pow(cos(a*t),exponent_z);
}

void algo2(int t, int r, float a, float b,
           int exponent_x, int exponent_y, int exponent_z,
           float *x, float *y, float *z)
{
    *x = r*(sin(a*t)*pow(cos(b*t),exponent_x));
    *y = r*(cos(a*t)*pow(cos(b*t),exponent_y));
    *z = r*pow(sin(a*t),exponent_z);
}

void algo3(int t, int r, float a, float b,
           int exponent_x, int exponent_y, int exponent_z,
           float *x, float *y, float *z)
{
    *x = r*(sin(a*t)*pow(sin(b*t),exponent_x));
    *y = r*(sin(a*t)*pow(cos(b*t),exponent_y));
    *z = r*pow(sin(a*t),exponent_z);
}

void algo4(int t, int r, float a, float b,
           int exponent_x, int exponent_y, int exponent_z,
           float *x, float *y, float *z)
{
    *x = r/4 * pow(a*sin(2*(t-PI/13)),exponent_x);
    *y = r/4 * pow(-b*cos(t),exponent_y);
    *z = r   * pow(sin(a*t),exponent_z);
}

void algo5(int t, int r, float a, float b,
           int exponent_x, int exponent_y, int exponent_z,
           float *x, float *y, float *z)
{
    *x = r*(sin(a*t)*pow(cos(a*t),exponent_x));
    *y = r*(sin(b*t)*pow(sin(b*t),exponent_y));
    *z = r*pow(sin(t),exponent_z);
}


/* --- Sort the circles on 'axis', from most distant to closest for simple
       "hidden line removal".  Farthest will be drawn first and overlapped
       by the nearer circles.
       Parms: queue = array of values needed sorting.
              index = single dimension array of integers indexing queue().
              left  = lowest index #.
              right = highest index #.
              axis  = string of either "X", "Y", or "Z"
*/
void quicksort(struct queue_type queue[], int *index, 
               int left, int right, char *axis)
{
    int i,j,
        *lstack, *rstack;

    if ((right-left) >= 1) {
        if (((lstack=malloc( 1+sizeof(int)*right )) == NULL) ||
            ((rstack=malloc( 1+sizeof(int)*right )) == NULL))
                error_fn(EMEM,"quicksort");

        median_index=index[(right+left)/2];
        median=component(&queue[median_index],axis);
        i=left-1; j=right+1;

        for(k=left; k <= right; k++) {
            if (component(&queue[index[k]],axis) < median) {  
                i++;
                lstack[i]=index[k];
            } else if (component(&queue[index[k]],axis) > median) {  
                j--;
                rstack[j]=index[k];
            }
        }
        for(k=left; k <= i; k++) {
            index[k]=lstack[k];
        }
        index[k]=median_index;
        for(k=j; k <= right; k++) {
            index[k]=rstack[k];
        }
        free(lstack);
        free(rstack);
        quicksort(queue, index, left, i, axis);
        quicksort(queue, index, j, right, axis);
    }
}


void clean_up_graph(void)
{
    closegraph();
}
